# Assetto Corsa freejoin server launcher
# by Keijo "Kegetys" Ruotsalainen, http://www.kegetys.fi

import http.server
import socketserver
import urllib.request
import time
import json
import sys
import configparser
import random
import subprocess
import os

RESERVE_TIME = 15 # How long slot is reserved for a joining player (time for player to connect)
LOCAL_TIMEOUT = 3 # Timeout in seconds for connecting to AC HTTP server
SERVER_TAG = "[FREEJOIN]" # Tag used to detect freejoin servers

def print(*args, **kwargs):
    __builtins__.print("[FREEJOIN] ", end="")
    return __builtins__.print(*args, **kwargs)

def get_server_setting(entry, default=None):
    ini = configparser.ConfigParser()
    ini.read("server/cfg/server_cfg.ini")
    try:
        return ini['SERVER'][entry]
    except:
        if default == None:
            print("ERROR: Could not read server_cfg file!")
            sys.exit()
        else:
            return default

num_entries = 0
def update_inis():
    # Update server settings
    global num_entries
    ini = configparser.ConfigParser()
    ini.optionxform = str
    ini.read("server/cfg/server_cfg.ini")
    
    # PICKUP_MODE_ENABLED apparently allows name change mid-race
    ini['SERVER']['PICKUP_MODE_ENABLED'] = "1"
    
    # Add tag to server name
    name = ini['SERVER']['NAME']
    if(SERVER_TAG not in name):
        ini['SERVER']['NAME'] = "%s %s" % (name, SERVER_TAG)
    
    with open("server/cfg/server_cfg.ini", 'w') as configfile:
        ini.write(configfile, space_around_delimiters=False)
        
    # Add GUIDs and driver names to entry list
    ini = configparser.ConfigParser()
    ini.optionxform = str
    ini.read("server/cfg/entry_list.ini")
    added = 0
    for idx,s in enumerate(ini.sections()):
        if(s[0:4] == "CAR_"):
            if idx < len(guids_list):
                ini.set(s,'GUID', str(guids_list[idx]))
                ini.set(s,'DRIVERNAME', '-')    
                added = added+1
            else:
                print("WARNING: No entry slot for client slot %d" % idx)
    if added == 0:
        print("ERROR: No valid entries in entry_list.ini!")
        sys.exit()
    with open("server/cfg/entry_list.ini", 'w') as configfile:
        ini.optionxform = str
        ini.write(configfile, space_around_delimiters=False)
    num_entries = added

def do_request(param):
    # Make request ot local server
    url = "http://localhost:%d/%s"%(PORT,param)
    req = urllib.request.Request(url=url,data=b'None',headers={'User-Agent':'Assetto Corsa Launcher'})
    return urllib.request.urlopen(req).read().decode("UTF-8")
    
def validate_guid(guid):
    # Check if GUID is valid
    try:
        #str = urllib.request.urlopen('http://localhost:%d/CANJOIN|%d' % (PORT, guid), None, LOCAL_TIMEOUT).read().decode("UTF-8")
        str = do_request("CANJOIN|%d"%guid)
        if str == "YES":
            return True
        else:
            return False
    except urllib.error.URLError:
        print("Could not connect to local AC server!")
        return False

def check_server():
    # Check if server is running
    try:
        #str = urllib.request.urlopen('http://localhost:%d/JSON|0' % (PORT), None, LOCAL_TIMEOUT).read().decode("UTF-8")
        str = do_request("JSON|0")
        if len(json.loads(str)['Cars']) > 0:
            return True
        else:
            return False
    except urllib.error.URLError:
        return False    
            
def server_in_race():
    # Check if server is in race mode
    try:
        #str = urllib.request.urlopen('http://localhost:%d/ENTRY' % (PORT), None, LOCAL_TIMEOUT).read().decode("UTF-8")
        str = do_request("ENTRY")
        return "Session: RACE" in str        
    except urllib.error.URLError:
        return False       
            
def slot_free(slot):
    # Check from server if slot is free
    try:        
        #str = urllib.request.urlopen('http://localhost:%d/JSON|0' % (PORT), None, LOCAL_TIMEOUT).read().decode("UTF-8")
        str = do_request("JSON|0")
        if json.loads(str)['Cars'][slot]['IsConnected']:
            print("Slot %d not free!" % slot)
            return False
        else:
            return True
    except urllib.error.URLError:
        print("Could not connect to local AC server!")
        return False

def guid_free(guid):
    # Check if guid is free
    if not guid in reserved_guids:
        return True
    if(reserved_guids[guid] > time.time()):
        print("GUID <%s> still reserved" % (guid))        
        return False
    return True
    
def reserve_guid(guid):
    # Reserve guid for player to give him time to join
    reserved_guids[guid] = time.time()+RESERVE_TIME    
    print("GUID <%s> reserved for %d seconds" % (guid, RESERVE_TIME))
    
MOTDtime = 0
MOTD = ""
def get_motd():
    # Read MOTD once every 30 seconds
    global MOTD, MOTDtime
    try:
        if motd_file == "":
            return ""
        if time.time()-MOTDtime > 0:
            with open(motd_file, 'r') as f:
                MOTD = f.read()
            MOTDtime = time.time()+30
        return MOTD
    except:
        return ""
     
def handle_motd(request, rHost, rPort):
    # Handler for message of the day
    # request contains user name and guid also here
    return get_motd()

def handle_reqjoin(request, rHost, rPort):
    # Handler for "REQJOIN"
    req = request.split('|')
    print("reqjoin %d" % len(req))
    if(len(req) < 5):
        return ""
    try:
        rPassword = req[1] # Requested password
        rId = int(req[2])  # Index of wanted slot
        rRealGuid = req[3] # Real GUID of player
        rRealName = req[4] # Player name
        
        error = None
        if(rPassword != PASSWORD and len(PASSWORD)>0):
            error = "Invalid password"        
        elif(rId<0 or rId >= len(guids_list)):
            error = "Invalid car ID"
        elif(not allow_race_join and server_in_race()):
            error = "Mid-race join disabled"
        
        if error == None:
            # Valid request, check GUID
            print("Slot request: %d from %s:%d (%s:%s)" % (rId, rHost, rPort, rRealName, rRealGuid))
            valid_guid = str(guids_list[rId])
            if not guid_free(valid_guid):
                error = "Slot reserved, refresh player list"
            elif not slot_free(rId):
                error = "Slot occupied, refresh player list"
            if error == None:
                # Send GUID to player
                reserve_guid(valid_guid) # GUID is reserved for a while
                return "OK|%s"%valid_guid
                
        # Wrong password or unjoinable slot
        print("Invalid request for slot %d, error: <%s> from %s:%d (%s:%s)" % (int(rId), str(error), rHost, rPort, rRealName, rRealGuid))
        return "NO|%s"%error
    except ValueError:            
        return ""
    
class rhandler(http.server.BaseHTTPRequestHandler):
    def log_message(self, format, *args):
        pass
        
    def do_GET(self):
        # Incoming HTTP request handler
        try:
            path = urllib.request.url2pathname(self.path)
            request = path.split('|',1)[0][1:]
            response = ""
            if request == "REQJOIN":
                response = handle_reqjoin(path, self.client_address[0], self.client_address[1])
            elif request == "MOTD":
                response = handle_motd(path, self.client_address[0], self.client_address[1])
            self.send_response(200)
            self.send_header('Content-type','text-html')
            self.end_headers()
            self.wfile.write(response.encode("utf-8"))            
        except:
            pass
   
 
# Read settings from server cfg
PORT = int(get_server_setting('HTTP_PORT'))
PASSWORD = get_server_setting('PASSWORD')
num_clients = int(get_server_setting('max_clients'))
allow_race_join = int(get_server_setting('ALLOW_RACE_JOIN', "1"))>0
motd_file = get_server_setting('MOTD', "")

# Create random GUIDs for players
guids_list = random.sample(range(1000, 100000000), num_clients)
reserved_guids = dict()

# Update config files
update_inis()

print("Server name <%s>" % get_server_setting('NAME'))
print("Clients: %d, HTTP port: %d, passsword: <%s>" % (num_clients, PORT, PASSWORD));

# Check server is not running
if check_server():
    print("ERROR: Server is already running!")
    sys.exit()

# Run the server, wait for it to init
if os.name == "posix":
    subprocess.Popen(["./acServer"])
else:
    subprocess.Popen(["./acServer.exe"])
time.sleep(4)
    
# Check that the local server is running
while not check_server():
    print("WARNING: Could not connect to AC server!")
    time.sleep(1)
print("Connected to local AC server")
   
# Validate guids list from server
for i in range(0, num_entries):    
    guid = guids_list[i]
    if not validate_guid(guid):
        print("ERROR: GUID <%s> is not valid!" % guid)
        sys.exit()
print("Validated %d GUIDs" % num_entries)
    
# Run HTTP server
serve_port = PORT+1
httpd = socketserver.TCPServer(("", serve_port), rhandler)
print("Serving at port", serve_port)
httpd.serve_forever()